// ******************************************************
//  file    HLsupport.cpp
//  purpose The class HLsupport is used by sdm.cpp to 
//			facilitate lookup/insertion/deletion of 
//			active locations
//  author  Alexandre Bouchard
//  written May 2003
// ******************************************************

// Note: First version, based on a hash

#include <cstdlib>

#include "HLsupport.h"
#include "sdm.h"

using namespace std;


/*
	Initialize the HLsupport with the given
	- number of dimensions dim
	- radius in each dimension radius
	- upper bounds and lower bounds in each dimension leftBounds and rightBounds
			respectively (the space is an hyperrectangle)
	
*/
			HLsupport::HLsupport(int dim, double* radius, double* leftBounds, double* rightBounds)
{
	numberOfDimensions = dim;
	Radius = new double[numberOfDimensions];
	LeftBounds = new double[numberOfDimensions];
	RightBounds = new double[numberOfDimensions];
	for (int i = 0; i < numberOfDimensions; i++)
	{
		Radius[i] = radius[i];
		LeftBounds[i] = leftBounds[i];
		RightBounds[i] = rightBounds[i];
	}
}

/*
	Destructor
*/
			HLsupport::~HLsupport()
{
	delete [] Radius;
	delete [] LeftBounds;
	delete [] RightBounds;
}

/*
	Add the active location of given index at given state in HLsupport ds.
	The insertion algorithm for the hashtable-based implementation is:
	0- Compute the key corresponding to the given state
	1- Find the list corresponding to key in the hashtable
	2a- If this list is NULL create a new list with given index in it and insert
				it in the hashtable in the entry corresponding to the given key.
				Note that in this case key is not deleted (see in the code below)
	2b- If the list is not NULL, insert the index in it and delete the key
*/
void		HLsupport::addHL(const State& state, int index)
{
	char* key = computeKey(state); // 0
	List* list = hashMap[key]; // 1
	if (list == NULL)
	{ // 2a
		list = new List();
		list->add(index);
		hashMap[key] = list;
	}
	else
	{ // 2b
		list->add(index);
		delete [] key;
	}
}

/*
	Remove the active location of given index at given state from HLsupport ds.
	The deletion algorithm for the hashtable-based implementation is:
	0- Compute the key corresponding to the given state
	1- Find the list corresponding to key in the hashtable
	2- Remove the index from the list
	3- If the list is empty, erase the corresponding entry in the hashtable,
				delete the original key used to store it and delete the list	 
*/
void		HLsupport::removeHL(const State& state, int index)
{
	char* key = computeKey(state); // 0
	List* list = hashMap[key]; // 1
	list->remove(index); // 2
	if (list->isEmpty())
	{ // 3
		const char* originalKey = hashMap.find(key)->first;
		hashMap.erase(key);
		delete list;
		delete [] originalKey;
	}
	delete [] key;
}

/*
	Given a state s, find a subset S of the inserted active locations A such that
	(for all x element of A) s is an element of x => x is an element of S
	(x can be viewed as an hyperball and s as a point)
	The objective is to return a set S as small as possible while keeping the running
	time for this method low.
	The lookup algorithm for the hashtable-based implementation is:
	0- Generates the set G(s) with getNeighborCubesRepresentatives(state)
	1- By assumption, (see comment before getNeighborCubesRepresentatives(state)),
				the subset of the space constituted of the union of the regions corresponding
				in the hash to the elements in G(s) covers the area activated by the given
				state (assuming uniform and fixed beta). It follows by metric symmetry that 
				the union of the lists of active locations corresponding in the hash to the 
				element in G(s) have the required property.
				So it is enough to compute the union of those lists, i.e.:
					1a- Compute the key of each element in G(s)
					1b- Find the list of corresponding elements in the hash (and delete the temp keys)
					1c- Carry out the partial unions
					1d- Delete G(s)
	TODO: prove (or disprove) that this complicated implementation is better than simple interval
					intersection (the previous implementation) (easy to show assuming uniform distribution
					but this is a bad assumption, actual tests might be the only way..)
				-> Tests on the mountain car domain suggest that this implementation is indeed better
	TODO: the implementation of List in SDM puts the elements in monotone order. This was useful in the
			context of the old implementation, in which intersections of lists had to be computed... but
			with the new implementation, it is just a waste of speed units
	REMINDER: THE CALLER IS RESPONSIBLE OF DELETING THE RETURNED LIST
*/
List*		HLsupport::potentiallyActiveSet(const State& state)
{
	ListOfStates* listOfNeighborCubesRepresentatives = getNeighborCubesRepresentatives(state); // 0
	List* listOfPotentiallyActiveSet = new List();
	ListOfStates::iterator listOfStatesIterator;
	for (listOfStatesIterator = listOfNeighborCubesRepresentatives->begin();
				listOfStatesIterator != listOfNeighborCubesRepresentatives->end();
				listOfStatesIterator++)
	{
		State* currentState = *listOfStatesIterator;
		char* key = computeKey(*currentState); // 1a
		List* currentList = hashMap[key]; // 1b
		delete currentState;
		if (currentList != NULL)
		{
			listOfPotentiallyActiveSet->listUnion(currentList); // 1c
		}
		else
		{
			// Subtle issue here: if there was no entry corresponding to key, one was created in
			// (1b) and deleting the key in (*) would then be problematic without the following
			// patch:
			hashMap.erase(key);
		}
		delete [] key; // (*)
	}
	delete listOfNeighborCubesRepresentatives; // 1d
	return listOfPotentiallyActiveSet;
}

/*
	(private method)
	Given a state, returns a key in the format of a string.
	Let P be the partition of the space into hyperrectangles obtained by taking the intersection of
	[the tiling of hypercube with sides of length 2*beta (the edge length of the active locations) s.t.
	the point (left[0], ... , left[N]) is the vertex of a hypercubic partition] , and [the space], where
	N is the number of dimension and "left" is the vector of lower bounds of the hypercubic space.
	Moreover we take the "left" sides of the rectangle to be closed and the "right" sides to be open
	in each dimension.
	So the algorithm is:
	0- Compute the index of the cell of P in which the given state sits. This can be done easily
				with the formula: ceiling((state.x[i] - left[i])/(2*beta[i]))
	1- We transform the obtained N-tuple of integer into a string in the following way:
				1a- Consider each integer in 256-ary representation and form a 4-tuple of char in the
							canonical way, then concatanate all those
				1b- Append \0
	Could be better, but does the job
	NOTE: we do not want non terminal zero's in the string, so the following modif is
	needed: Add one more char in the string at the end with value one.
	For all nonterminal key[.] with value zero, add one to key[.], count the number of
	such zero key[.] and add this number to the last char of the string.
	REMINDER: THE CALLER IS RESPONSIBLE OF DELETING THE RETURNED STRING
*/
char*		HLsupport::computeKey(const State& state)
{
	char* key = new char[4*numberOfDimensions + 2]; // 4 for each dimension + zero occurrence (see above) + \0
	int zeroOccurrences = 0;
	for (int i = 0; i < numberOfDimensions; i++)
	{
		// since (state.x[i] - left[i]) / (2*radius[i]) is positive, we can do step 0 with the following cast:
		int currentPartitionIndex = (int) ( (state.x[i] - LeftBounds[i]) / (2*Radius[i]) ); // 0
		// 1a
		key[4*i] = (char) currentPartitionIndex; // work because it copies the lower bits
		if (key[4*i] == 0)
		{
			zeroOccurrences++;
			(key[4*i])++;
		}
		currentPartitionIndex >>= 8; // shift right, discards the bits already considered
		key[4*i+1] = (char) currentPartitionIndex;
		if (key[4*i + 1] == 0)
		{
			zeroOccurrences++;
			(key[4*i + 1])++;
		}	
		currentPartitionIndex >>= 8;
		key[4*i+2] = (char) currentPartitionIndex;
		if (key[4*i + 2] == 0)
		{
			zeroOccurrences++;
			(key[4*i + 2])++;
		}
		currentPartitionIndex >>= 8;
		key[4*i+3] = (char) currentPartitionIndex;
		if (key[4*i + 3] == 0)
		{
			zeroOccurrences++;
			(key[4*i + 3])++;
		}	
	}
	key[4*numberOfDimensions] = 1 + zeroOccurrences;
	key[4*numberOfDimensions + 1] = '\0'; // 1b
	return key;
}

/*
	(private method)
	Given a state s, generate the minimal set G(s) s.t. the union of the regions (k^(-1))(k(t))
	when t ranges over G(s) covers the HL centered at s with radius beta. [where k(.) = computeKey(.)]
	It can be proven to be equivalent to computing the position of the 2^N vertices of the hyperrectangle, 
	but for each edges, the left points that are not in the space are not considered (because the k(t) is not
	defined for them) (the right points that are not in the space are still considered since the right points
	of the partition cells are open)
	
	We show it works by induction on the dimensionality D of the space:
	- The base case (D = 1) is trivial:
	The right end of the interval is added (1a) and the left end of the interval is added iff it is in the
	space (1b)
	- Suppose that it works up to D - 1:
	It can be proven that the set of vectices S1 corresponding to the left (resp. right, S2) end-points of the edges
	along the D-th dimension form a (D-1)-dimensional hypercube. So it is enough to find the vertices of the
	hypercube corresponding to S1 (2a) and to add those of the hypercube corresponding to S2 if they are in the space
	(2b). It can be done using getNeighborCubesRepresentatives by the induction hypothesis
	
	NOTE: we assume:- Closed, hyperrectangular space
									- Partitions as described and produced in computeKey()
									- Hyperrectangular active locations of fixed and uniform radius beta
	REMINDER: THE CALLER IS RESPONSIBLE OF DELETING THE RETURNED LIST ***AND ITS CONTENTS***
*/
ListOfStates*   HLsupport::getNeighborCubesRepresentatives(	const State& state, int currentDimension)
{
	if (currentDimension == -1)
	{
		// The default value is the numberOfDimension-th dimension
		currentDimension = numberOfDimensions - 1;
	}
	if (currentDimension < 0)
	{
		cout << "Invalid number of dimensions: " << currentDimension << "\n";
		cout << "Dimensionality of space: " << numberOfDimensions;
		exit(EXIT_FAILURE);
	}
	ListOfStates* returnedList;
	if (currentDimension == 0)
	{
		returnedList = new ListOfStates();
		State* state1 = new State(state);
		state1->x[0] = state.x[0] + Radius[0];
		returnedList->push_front(state1); // 1a
		if ((state.x[0] - Radius[0]) > LeftBounds[0])
		{
			State* state2 = new State(state);
			state2->x[0] = state.x[0] - Radius[0];
			returnedList->push_front(state2); // 1b
		}
	}
	else
	{
		State* state1 = new State(state);
		state1->x[currentDimension] = state.x[currentDimension] + Radius[currentDimension];
		returnedList = getNeighborCubesRepresentatives(*state1, currentDimension - 1); // 2a
		if ((state.x[currentDimension] - Radius[currentDimension]) > LeftBounds[0])
		{
			State* state2 = new State(state);
			state2->x[currentDimension] = state.x[currentDimension] - Radius[currentDimension];
			ListOfStates* temp = getNeighborCubesRepresentatives(*state2, currentDimension - 1);
			ListOfStates::iterator tempIterator;
			for (tempIterator = temp->begin();
						tempIterator != temp->end();
						tempIterator++)
			{
				returnedList->push_front(*tempIterator);	// 2b	
			}
			delete temp;
		}
	}
	return returnedList;	
}




















// TESTS
// To run them, uncomment the fct above (and also the declaration in ~.h


/*
	Execute a sequence of test cases on HLsupport

int main()
{
	cout << "Test cases launched..." << endl;
	//testHashTable();
	//testComputeKey(); // see testComputeKey before uncommenting
	//testGetNeighborCubesRepresentatives(); // see testGetNeighborCubesRepresentatives before uncommenting
	testAddHL();
	cout << "Test completed" << endl;
} */

/*
	Test the SGI hashtable extension of STL under heavy load
*/
void	testHashTable()
{
	HashMap hashTable;
	
	// insert 200 000 entries
	for (int j = 2; j < 1000; j++)
	{
		for (int i = 0; i < 200; i++)
		{
			char* key = new char[j];
			for (int k = 0; k < j - 1; k++)
			{
				key[k] = i;
			}
			key[j - 1] = '\0';
			List* currentList = new List();
			currentList->add(i);
			hashTable[key] = currentList;
		}
	}
	cout << endl;
	
	// check they are still there
	for (int j = 2; j < 1000; j++)
	{
		for (int i = 0; i < 200; i++)
		{
			char* key = new char[j];
			for (int k = 0; k < j - 1; k++)
			{
				key[k] = i;
			}
			key[j - 1] = '\0';
			List* checkList = hashTable[key];
			if (checkList == NULL)
			{
				cout << "testHashTable..[FAILED] (insertion in hash failed)" << endl;
				return;
			}
			else
			{
				if (checkList->isEmpty())
				{
					cout << "testHashTable..[FAILED] (insertion in list failed)" << endl;
					return;
				}
			}
		}
	}
	cout << "testHashTable..[ok]" << endl;
}

/*
	Test the method HLsupport::addHL()
*/
void testAddHL()
{
	// 1- Basic test. Just checks that it does not crash
	double* radius = new double[3];
	double* left = new double[3];
	double* right = new double[3];
	radius[0] = 0.2; radius[1] = 0.2; radius[2] = 0.2;
	left[0] = -5; left[1] = -5; left[2] = -5;
	right[0] = 5; right[1] = 5; right[2] = 5;
	HLsupport* support1 = new  HLsupport(3, radius, left, right);
	State state;
	state.dimensionality = 3;
	double* x = new double[3];
	x[0] = 1.1; x[1] = -3.4; x[2] = 2.7;
	state.x = x;
	support1->addHL(state, 1);
	delete [] radius;
	delete [] left;
	delete [] right;
	delete [] x;
	delete support1;
	
	// 2- Test massive insertion
	radius = new double[3];
	left = new double[3];
	right = new double[3];
	radius[0] = 0.2; radius[1] = 0.2; radius[2] = 0.2;
	left[0] = -50; left[1] = -50; left[2] = -50;
	right[0] = 50; right[1] = 50; right[2] = 50;
	HLsupport* support2 = new  HLsupport(3, radius, left, right);	
	State* randomState;
	for (int i = 0; i < 500; i++)
	{
		randomState = new State();
		randomState->dimensionality = 3;
		x = new double[3];
		randomState->x = x;
		x[0] = ((double) random()) / ((double) RAND_MAX) * ((double) 100) - 50;
		x[1] = ((double) random()) / ((double) RAND_MAX) * ((double) 100) - 50;
		x[2] = ((double) random()) / ((double) RAND_MAX) * ((double) 100) - 50;
		printStringAsIntegers(support2->computeKey(*randomState)); cout << endl;
		support2->addHL(*randomState, i);
		
		delete [] x;
		delete randomState;
	}
	cout << endl;
	delete [] radius;
	delete [] left;
	delete [] right;
	delete support2;	
	
	cout << "testGetAddHL..[ok]" << endl;
}

/*
	Debugging function... prints a string as a space-separated list of integers
*/
void	printStringAsIntegers(char* string)
{
	for (int i = 0; string[i] != '\0'; i++)
	{
		int temp = (unsigned char) string[i];
		cout << temp << " ";
	}
}

void	printCoordinates(const State& state)
{
	cout << "(";
	for (int i = 0; i < state.dimensionality; i++)
	{
		cout << (state.x)[i] << ",";
	}
	cout << ")";
}

/*
	Test the method HLsupport::computeKey
	NOTE: The method has to be changed to public before running the test
	Also uncomment the call in the test
*/
void testComputeKey()
{
	// 1- Basic test. Just checks that it does not crash
	double* radius = new double[3];
	double* left = new double[3];
	double* right = new double[3];
	radius[0] = 0.2; radius[1] = 0.2; radius[2] = 0.2;
	left[0] = -5; left[1] = -5; left[2] = -5;
	right[0] = 5; right[1] = 5; right[2] = 5;
	HLsupport* support1 = new  HLsupport(3, radius, left, right);
	State state;
	state.dimensionality = 3;
	double* x = new double[3];
	x[0] = 1.1; x[1] = -3.4; x[2] = 2.7;
	state.x = x;
	//support1->computeKey(state);
	delete [] radius;
	delete [] left;
	delete [] right;
	delete [] x;
	delete support1;
	
	// 2- Random generation/collision test
	// TODO: generate random point in space with medium to high
	//		dimensionality/size so that the probability of collision is low.
	//		Verifies that indeed there is no collision
	
	cout << "testComputeKey...[ok]" << endl;
}

/*
	Test the method HLsupport::getNeighborCubesRepresentatives()
	NOTE: The method has to be changed to public before running the test
	Also uncomment the call in the test
*/
void testGetNeighborCubesRepresentatives()
{
	// 1- Basic test. Just checks that it does not crash
	double* radius = new double[3];
	double* left = new double[3];
	double* right = new double[3];
	radius[0] = 0.2; radius[1] = 0.2; radius[2] = 0.2;
	left[0] = -5; left[1] = -5; left[2] = -5;
	right[0] = 5; right[1] = 5; right[2] = 5;
	HLsupport* support1 = new  HLsupport(3, radius, left, right);
	State state;
	state.dimensionality = 3;
	double* x = new double[3];
	x[0] = 0; x[1] = 0; x[2] = 0;
	state.x = x;
	/*ListOfStates* listOfNeighborCubesRepresentatives = support1->getNeighborCubesRepresentatives(state);	
	ListOfStates::iterator i;
	for (i = listOfNeighborCubesRepresentatives->begin();
				i != listOfNeighborCubesRepresentatives->end();
				i++)
	{
		State* currentState = *i;
		cout << "(" << currentState->x[0] << ", " << currentState->x[1] << ", " << currentState->x[2] << ")" << endl;
		delete currentState;
	}
	delete listOfNeighborCubesRepresentatives;
	delete [] radius;
	delete [] left;
	delete [] right;
	delete [] x;
	delete support1;*/
	
	// 2- Behavior in boundary conditions
	// TODO
	
	// 3- Test that G(s) is s.t. the union of the regions (k^(-1))(k(t))
	//			when t ranges over G(s) covers the HL centered at s with radius beta.
	// TODO
	
	cout << "testGetNeighborCubesRepresentatives...[ok]" << endl;
}
