#include <stddef.h>
#include <iostream.h>
#include <fstream.h>

#include "interface_classes.h"
#include "sdm.h"


//----------------------------------------------------


SDM::SDM(sdmParameters& sdmP, double (*simFunction)(const State& s1, const State& s2, double* threshold)){
  
  int i;

  Capacity = sdmP.memCapacity;
  similarity = simFunction;
  
  ActivationRadius = new double[State::dimensionality];
  for (i=0; i<State::dimensionality; i++)
    ActivationRadius[i] = sdmP.radius[i];

  SufficientSamples = sdmP.SufficientSamples;
  UsedCapacity = 0;
  InitialAlpha = 0.15;
  AlphaSchedule = CONSTANT_ALPHA;
  AlphaDecreaseConst = 1.5;
  UseTraces = false;
  Rehearsing = 0;
  RememberRehearsing = 0.0;
  Time = 0;
  RandomAdditionsFactor=0;
  GetInactiveAttempt = 0;
  MaximumSamples = 3*SufficientSamples;
  SuttonChanges=0;
  Threshold =sdmP.Threshold;
  infFile = NULL;
  inf.Added=0;
  inf.Removed=0;
  inf.Moved=0;
  inf.Scanned=0;
  inf.Activated=0;
  inf.Samples=0;


  HardArray = new HardLocation[Capacity];
  if (HardArray == NULL){
    cout << "Memory allocation problem (SDM constructor, HardArray)" << endl;
    exit(EXIT_FAILURE);
  }

  support = new HLsupport(State::dimensionality, ActivationRadius, sdmP.left, sdmP.right);

  if (support == NULL){
    cout << "Memory allocation problem (SDM constructor, support)" << endl;
    exit(EXIT_FAILURE);
  } 

  //distribute some locations randomly form the start

  if (SufficientSamples <= 0){
    cout << "Specify the minimum desired number of hard locations." << endl;
    exit(EXIT_FAILURE);
  }
  else{

    int j;
    State s;
    double Weight;
    List* potActiveSet;
    Node* hl;
    int NumClose;

    if (Threshold < 0){
      if (SufficientSamples == 1)
	Threshold = 0.1;
      else 
	if (SufficientSamples == 2)
	  Threshold = 0.55;
	else
	  Threshold = 1.0-1.0/(double)(SufficientSamples-1)+0.05;
    }
    

    for (i=0; i<sdmP.StartingSize*10; i++){
      if (UsedCapacity == sdmP.StartingSize) break;
      for (j=0; j<State::dimensionality; j++){
	s.x[j] = sdmP.left[j] + (rand()/(double)RAND_MAX)*(sdmP.right[j]-sdmP.left[j]);
      }

      NumClose=0;
      potActiveSet = support->potentiallyActiveSet(s);
      if (potActiveSet != NULL){
	hl = potActiveSet->walk(start);
	while (hl != NULL){
	  Weight = (*similarity)(HardArray[hl->index].state, s, ActivationRadius);
	  if (Weight > Threshold)
	    
	    NumClose++;
	  hl = potActiveSet->walk(continue);
	}
      }
      if (NumClose < 1){

	support->addHL(s,UsedCapacity);
	HardArray[UsedCapacity].state = s;
	HardArray[UsedCapacity].value = DEFAULT_VALUE;
	HardArray[UsedCapacity].alpha = InitialAlpha;
	HardArray[UsedCapacity].EligibilityTrace = 1;
	HardArray[UsedCapacity].ActivationCounter = 1;
	HardArray[UsedCapacity].TimeStamp = Time;
	UsedCapacity++;
      }
    }
  }
}

//=================================================

void SDM::localActive(const State& s, NeighborhoodInfo& ni){
  
  double Weight;
  List* potActiveSet;
  Node* hl;
  int r, i;

  ni.Prediction = 0;
  ni.NumClose=0;
  ni.MaxWeight = 0.0;
  ni.MaxIndex = -1;
  ni.Activated = 0;
  ni.WeightSum = 0;
  ni.RandIndex = -1;

  potActiveSet = support->potentiallyActiveSet(s);
  if (potActiveSet != NULL){

    
    hl = potActiveSet->walk(start);
    
    while (hl != NULL){
      Weight = (*similarity)(HardArray[hl->index].state, s, ActivationRadius);
      if (Weight > 0.00001){

	ni.Prediction+=Weight*HardArray[hl->index].value;
	ni.WeightSum+=Weight;
	
	if (Weight > ni.MaxWeight){
	  ni.MaxWeight = Weight;
	  ni.MaxIndex = hl->index;
	}
	if (Weight > Threshold)
	  ni.NumClose++;
	
	ni.Activated++;
      }
      hl = potActiveSet->walk(continue);
      if ((Weight < 0.00001) && (hl != NULL)) {
	potActiveSet->remove(hl->prev);
      }
    }

    if (ni.Activated > 0){
      r = rand() % ni.Activated;
      hl = potActiveSet->walk(start);
      
      for (i=0; i<r; i++){
	hl = potActiveSet->walk(continue);    
      }
      ni.RandIndex = hl->index;
      ni.RandWeight = (*similarity)(HardArray[hl->index].state, s, ActivationRadius);
    }
  }

  if ((ni.Activated == 0 ) || (ni.WeightSum < 0.0001))
    ni.Prediction = DEFAULT_VALUE;
  else
    ni.Prediction /=  ni.WeightSum;

  delete potActiveSet;
}


//==================================================

double SDM::localPredict(const State& s){
  
  double Weight, SampleValue;
  List* potActiveSet;
  Node* hl;
  int Activated = 0;
  double WeightSum = 0.0;
  double output = 0;
  
  potActiveSet = support->potentiallyActiveSet(s);
  if (potActiveSet == NULL){
    return DEFAULT_VALUE;
  }

  hl = potActiveSet->walk(start);
  while (hl != NULL){
    Weight = (*similarity)(HardArray[hl->index].state, s, ActivationRadius);
    if (Weight > 0){
      
      output+=Weight*HardArray[hl->index].value;
      WeightSum+=Weight;
      Activated++;
    }
    hl = potActiveSet->walk(continue);
  }


  if ((Activated == 0 ) || (WeightSum < 0.0001))
    output = DEFAULT_VALUE;
  else
    output = output / WeightSum;

  
  delete potActiveSet;
  return output;
  
  }

//========================================================

void SDM::localUpdate(const State& s, double target){

  List* potActiveSet;
  Node* hl;
  double WeightSum=0;
  double Prediction=0;
  double Weight;

  potActiveSet = support->potentiallyActiveSet(s);
  if (potActiveSet != NULL){
    hl = potActiveSet->walk(start);
    while (hl != NULL){
      Weight = (*similarity)(HardArray[hl->index].state, s, ActivationRadius);
      inf.Scanned++;
      if (Weight > 0.00001){
	inf.Activated++;
	Prediction+=Weight*HardArray[hl->index].value; 
	WeightSum += Weight;
      }

      hl = potActiveSet->walk(continue);
      if ((Weight < 0.00001) && (hl != NULL)) {
	potActiveSet->remove(hl->prev);
      }
    }
  }
    

  if (WeightSum>0){
    
    Prediction = Prediction/WeightSum;
    hl = potActiveSet->walk(start);

    while (hl != NULL){
      Weight = (*similarity)(HardArray[hl->index].state, s, ActivationRadius);
      HardArray[hl->index].value = HardArray[hl->index].value + (target - Prediction)*HardArray[hl->index].alpha*Weight/WeightSum;
      if (AlphaSchedule == DECREASE_ALPHA)
	HardArray[hl->index].alpha = AlphaDecreaseConst/(AlphaDecreaseConst/HardArray[hl->index].alpha+1);
	
      hl = potActiveSet->walk(continue);
    }
  }
  delete potActiveSet;
}
//===========================================================

void SDM::removeHL(int index){

  NeighborhoodInfo ni;
  support->removeHL(HardArray[index].state, index);

  localActive(HardArray[index].state, ni);
  
  
  
  if (ni.Activated > 0){
    support->removeHL(HardArray[ni.MaxIndex].state,ni.MaxIndex);
    int i;
    for (i=0; i<State::dimensionality; i++){
      HardArray[ni.MaxIndex].state.x[i] = (HardArray[ni.MaxIndex].state.x[i] + HardArray[index].state.x[i])/2.0;
    }
    support->addHL(HardArray[ni.MaxIndex].state,ni.MaxIndex);
    HardArray[ni.MaxIndex].value = (HardArray[ni.MaxIndex].value + HardArray[index].value)/2; 
  }

  if (index != (UsedCapacity-1)){
    support->removeHL(HardArray[UsedCapacity-1].state,UsedCapacity-1);
    support->addHL(HardArray[UsedCapacity-1].state,index);
    HardArray[index].state = HardArray[UsedCapacity-1].state;
    HardArray[index].value = HardArray[UsedCapacity-1].value;
    HardArray[index].alpha = HardArray[UsedCapacity-1].alpha;
    HardArray[index].EligibilityTrace = HardArray[UsedCapacity-1].EligibilityTrace;
    HardArray[index].ActivationCounter = HardArray[UsedCapacity-1].ActivationCounter;
    HardArray[index].TimeStamp = HardArray[UsedCapacity-1].TimeStamp;
  }
  UsedCapacity--;
}

//============================================================

void SDM::predict(const State& s, double& output){
  
  Time++;
  NeighborhoodInfo ni;
  localActive(s,ni);
     
  if (ni.Activated < SufficientSamples){
    ni.Prediction = (ni.Prediction + DEFAULT_VALUE*(SufficientSamples-ni.Activated)) / (ni.WeightSum + (SufficientSamples-ni.Activated));
    addHL(s,ni);
    
  }

  output = ni.Prediction;

  

}

//===============================================

void SDM::addHL(const State& s, NeighborhoodInfo& ni){
 

if (UsedCapacity < Capacity) {

    if ((ni.MaxWeight < Threshold) || ((ni.Activated == 1) && (RandomAdditionsFactor == 0)) ){//add this sample as a new HL){
	
      support->addHL(s,UsedCapacity);
      HardArray[UsedCapacity].state = s;
      HardArray[UsedCapacity].value = ni.Prediction;
      HardArray[UsedCapacity].alpha = InitialAlpha;
      HardArray[UsedCapacity].EligibilityTrace = 1;
      HardArray[UsedCapacity].ActivationCounter = 1;
      HardArray[UsedCapacity].TimeStamp = Time;
      UsedCapacity++;
      ni.Activated++;
      inf.Added++;
    }
    else{
      
      bool replace=false; 
      if ((ni.Activated == 2) && (RandomAdditionsFactor == 0)){
	int secondIndex;
	double Weight, secondWeight;
	List* potActiveSet;
	Node* hl;

	potActiveSet = support->potentiallyActiveSet(s);
	hl = potActiveSet->walk(start);
	if (hl->index == ni.MaxIndex)
	  hl = potActiveSet->walk(continue);

	Weight = (*similarity)(HardArray[hl->index].state, s, ActivationRadius);
	secondWeight = (*similarity)(HardArray[hl->index].state, HardArray[ni.MaxIndex].state, ActivationRadius);
	    
	if (Weight < secondWeight){
	  replace = true;
	  inf.Moved++;
	}
	delete potActiveSet;
      }
      
      if (((ni.Activated > 2)&&(ni.NumClose < 2)) || (replace == true)){
	support->removeHL(HardArray[ni.MaxIndex].state,ni.MaxIndex);
	support->addHL(s,ni.MaxIndex);
	HardArray[ni.MaxIndex].state = s;
	HardArray[ni.MaxIndex].value = ni.Prediction; 
	HardArray[ni.MaxIndex].alpha = InitialAlpha;
	HardArray[ni.MaxIndex].EligibilityTrace = 1;
	HardArray[ni.MaxIndex].ActivationCounter = 1;
	HardArray[ni.MaxIndex].TimeStamp = Time;
	
      }
      
    }
      
    if ((ni.Activated < SufficientSamples) && (RandomAdditionsFactor > 0)){

      State sNew;
      int ToAdd = SufficientSamples - ni.Activated;
      bool success;
      double Prediction;

      while ((ToAdd > 0) &&(UsedCapacity < Capacity)){
	randomSample(s, RandomAdditionsFactor, sNew, success, Prediction);
	if (success == true){ 
	  support->addHL(sNew,UsedCapacity);
	  HardArray[UsedCapacity].state = sNew;
	  HardArray[UsedCapacity].value = Prediction;
	  HardArray[UsedCapacity].alpha = InitialAlpha;
	  HardArray[UsedCapacity].EligibilityTrace = 1;
	  HardArray[UsedCapacity].ActivationCounter = 1;
	  HardArray[UsedCapacity].TimeStamp = Time;
	  UsedCapacity++;
	  inf.Added++;
	  ni.Activated++;
	  
	}
	ToAdd--;
      }

    }
  }

  if ( (UsedCapacity == Capacity) && (GetInactiveAttempt > 0)){
    int ind, i;
    int j;

    NeighborhoodInfo pi;
    double Weight = 1.1;
   
    int Act=ni.Activated;

    for (j=0; j< (SufficientSamples - Act); j++){
      
      for (i=0; i<GetInactiveAttempt; i++){
	ind = rand() % Capacity;
	Weight = (*similarity)(HardArray[ind].state, s, ActivationRadius);
	if (Weight < 0.0001)
	  break;
      }



      if (Weight < 0.0001){
	//cout << "point 2" << endl;
	SuttonChanges++;

	removeHL(ind);
	ind = UsedCapacity;
	inf.Moved++;

	if (ni.MaxWeight < Threshold){
	  support->addHL(s,ind);
	  HardArray[ind].state = s;
	  HardArray[ind].value = ni.Prediction;
	  HardArray[ind].alpha = InitialAlpha;
	  HardArray[ind].EligibilityTrace = 1;
	  HardArray[ind].ActivationCounter = 1;
	  HardArray[ind].TimeStamp = Time;
	  ni.Activated++;
	  
	  UsedCapacity++;
	}
	else{

	  bool success;
	  State sNew;
	  double Prediction;
	  int k;
	  for (k=0; k<RandomAdditionsFactor; k++){
	    randomSample(s, RandomAdditionsFactor, sNew, success, Prediction);
	    if (success == true){
	      inf.Moved++;
	      support->addHL(sNew,ind);
	      HardArray[ind].state = sNew;
	      HardArray[ind].value = Prediction;
	      HardArray[ind].alpha = InitialAlpha;
	      HardArray[ind].EligibilityTrace = 1;
	      HardArray[ind].ActivationCounter = 1;
	      HardArray[ind].TimeStamp = Time;
	      ni.Activated++;
	      UsedCapacity++;
	      break;
	    }
	  }
	
	}
      }
    }
  }// UsedCapacity==Capacity
  
}
  
//-----------------------------------------------------


void SDM::learn(const State& s, const double target){

  if (UseTraces == true) learnWithTraces(s, target);
  if (UseTraces == false) learnNoTraces(s, target);

}

void SDM::learnWithTraces(const State& s, double target){}

//================================================================

void SDM::learnNoTraces(const State& s, double target){

  NeighborhoodInfo ni;
  inf.Samples++;
  
  localActive(s, ni);
  
  if (ni.Activated > MaximumSamples){
    removeHL(ni.RandIndex);
    ni.Activated--;
    inf.Removed++;
    
  }

  if (ni.Activated < SufficientSamples){
    ni.Prediction = target;
    addHL(s, ni);
  }
    

  localUpdate(s, target);

  

}
//====================================

void SDM::randomSample(const State& s, int Attempts, State& sample, bool& success, double& output){

  int i,j;
  double WeightA, WeightSum;
  List* potActiveSetA;
  Node* hlA;
  int NumClose;
  success = false;
  State sNew;

  i=0;
  while ((i<Attempts) && (success == false)){
    i++;
    for (j=0; j<State::dimensionality; j++){
      sNew.x[j] = (s.x[j] - ActivationRadius[j])+2.0*ActivationRadius[j]*rand()/(double)(RAND_MAX);
    }

    NumClose=0;
    output=0;
    WeightSum=0;
    potActiveSetA = support->potentiallyActiveSet(sNew);
    if (potActiveSetA != NULL){
      hlA = potActiveSetA->walk(start);
      while (hlA != NULL){
	WeightA = (*similarity)(HardArray[hlA->index].state, sNew, ActivationRadius);
	if (WeightA > Threshold){
	  NumClose++;
	  break;
	}
	if (WeightA > 0){
	  output += WeightA*HardArray[hlA->index].value;
	  WeightSum+=WeightA;
	}
	  
	hlA = potActiveSetA->walk(continue);
      }
    }
    delete potActiveSetA;
    if (NumClose < 1){
      success = true;
      sample = sNew;
      if (WeightSum > 0)
	output = output/WeightSum;
      else output=DEFAULT_VALUE;
    }
  }
}
//=================================================

void SDM::randomOutSample(const State& s, int Attempts, State& sample, bool& success){

  int i,j;
  double WeightA;
  List* potActiveSetA;
  Node* hlA;
  int NumClose;
  success = false;
  State sNew;

  i=0;
  while ((i<Attempts) && (success == false)){
    i++;
    for (j=0; j<State::dimensionality; j++){
      sNew.x[j] = support->LeftBounds[j] + (rand()/(double)RAND_MAX)*(support->RightBounds[j]-support->LeftBounds[j]);
    }

    NumClose=0;
    potActiveSetA = support->potentiallyActiveSet(sNew);
    if (potActiveSetA != NULL){
      hlA = potActiveSetA->walk(start);
      while (hlA != NULL){
	WeightA = (*similarity)(HardArray[hlA->index].state, sNew, ActivationRadius);
	if (WeightA > Threshold){
	  NumClose++;
	  break;
	}
	hlA = potActiveSetA->walk(continue);
      }
    }
    delete potActiveSetA;
    if (NumClose < 1){
      WeightA = (*similarity)(s, sNew, ActivationRadius);
      if (WeightA < 0.0001){
	success = true;
	sample = sNew;
      }
    }
  }
}


//-----------------------------------------------------  
void SDM::computeGradient(const State& s, double* GradientVector){

}
  /* Compute the gradient w.r.t. architecture parameters at the current parameters' values and input s
   */

void SDM::updateParameters(double* delta){

}
  /* Update parameters by amounts in delta array, 
     possible multiplied by appropriate learning steps.
  */

void SDM::replaceTraces(const State& s, double replace){

}
  /* Replace traces of parameters, activated by input state s to value replace.
   */

void SDM::decayTraces(double factor){

}
  /* Decay (multiply) traces by factor.
   */

void SDM::accumulateTraces(const State& s, double amount){

}
  /* Increment traces by amount for parameters activated by input s.
   */

//-----------------------------------------------
void SDM::setArchitectureParameters(int argc, char *argv[]){

  int i, j;
  char buffer[80];
  char c;

  if (argc==0){
    cout << "No file name to load data." << endl;
    exit(EXIT_SUCCESS);
  }
  ifstream file(argv[0]);
  if (file.fail()){
    cout << "Error (table): cannot open file " << argv[0] << endl;
    exit(EXIT_FAILURE);
  }

  file.get(buffer,strlen("Size: ")+1);
  if (file.fail()){
    cout << "Error (table): input failed 1" << endl;
    exit(EXIT_FAILURE);
  }

  file >> UsedCapacity;
  file.get(c);

  for(i=0; i<UsedCapacity; i++){
    for (j=0; j<State::dimensionality; j++){
      file >> HardArray[i].state.x[j];
      file.get(c);
      if (file.fail()){
	cout << "Error 1 (SDM): error on read from "<< argv[0] << endl;
	exit(EXIT_FAILURE);
      }
    }
    file >> HardArray[i].value;
    HardArray[i].alpha = InitialAlpha;
    HardArray[i].EligibilityTrace = 1;
    HardArray[i].ActivationCounter = 1;
    HardArray[i].TimeStamp = Time;
    support->addHL(HardArray[i].state,i);

    file.get(c);
    if (file.fail()){
      cout << "Error 2 (SDM): error on read from "<< argv[0] << endl;
      exit(EXIT_FAILURE);
    }
  }
  file.close();

}
  


//--------------------------------------------------

void SDM::saveArchitectureParameters(int argc, char *argv[]){

  int i, j;
  if (argc==0){
    cout << "No file name to save data." << endl;
    exit(EXIT_SUCCESS);
  }
  ofstream file(argv[0]);
  if (file.fail()){
    cout << "Error (SDM): cannot open file " << argv[0] << endl;
    exit(EXIT_FAILURE);
  }

  file << "Size: " << UsedCapacity << endl;
  if (file.fail()){
    cout << "Error (SDM): error on write to "<< argv[0] << endl;
    exit(EXIT_FAILURE);
  }

  for(i=0; i<UsedCapacity; i++){
    for (j=0; j<State::dimensionality; j++){
      file << HardArray[i].state.x[j] << " ";
      if (file.fail()){
	cout << "Error (SDM): error on write to "<< argv[0] << endl;
	exit(EXIT_FAILURE);
      }
    }
    file << HardArray[i].value << endl;
    if (file.fail()){
      cout << "Error (SDM): error on write to "<< argv[0] << endl;
      exit(EXIT_FAILURE);
    }
  }
 
  file.close();
}
  
//------------------------------------------------

void SDM::setLearningParameters(int argc, char *argv[]){

  int i;
  
  for (i=0; i<argc; i++){
    

    if (strncmp("schedule=", argv[i], 9)==0){
      if (strncmp("schedule=constant", argv[i], 17)==0)
	AlphaSchedule=CONSTANT_ALPHA;
      if (strncmp("schedule=decrease", argv[i], 17)==0)
	AlphaSchedule=DECREASE_ALPHA;
    }

    if (strncmp("alpha=", argv[i], 6)==0){
      InitialAlpha=atof(&(argv[i][6]));
      if ((InitialAlpha<=0) || (InitialAlpha>1)){
	cout << "Error (SDM): learning step must be in (0,1]" << endl;
	exit(EXIT_FAILURE);
      }
      
    }

    if (strncmp("d=", argv[i], 2)==0){
      AlphaDecreaseConst=atof(&(argv[i][2]));
    }

    if (strncmp("t=", argv[i], 2)==0){
      UseTraces=atoi(&(argv[i][2]));
    }

    if (strncmp("ra=", argv[i], 3)==0){
      RandomAdditionsFactor=atoi(&(argv[i][3]));
    }

    if (strncmp("gi=", argv[i], 3)==0){
      GetInactiveAttempt=atoi(&(argv[i][3]));
    }

    if (strncmp("ms=", argv[i], 3)==0){
      MaximumSamples=atoi(&(argv[i][3]));
    }


    if (strncmp("reh=", argv[i], 4)==0){
      Rehearsing=atof(&(argv[i][4]));
      RememberRehearsing = Rehearsing;
    }

    if (strncmp("reht=", argv[i], 5)==0){
      RememberRehearsing = Rehearsing;
      Rehearsing=atof(&(argv[i][5]));
    }

    if (strncmp("reho", argv[i], 5)==0){
       Rehearsing = RememberRehearsing;
      
    }
    
  }

}
  
//=======================================

void SDM::setID(int id){
  ID=id;
  char* infFileName = new char[20];
  strcpy(infFileName, "stat");
  char temp[5];
  sprintf(temp,".a%d",ID);
  strcat(infFileName,temp);
  infFile = new ofstream(infFileName);
  delete infFileName;
}

//=====================================

int SDM::getSize(){
  
  
  if ((infFile != NULL) && (inf.Samples >0))
    (*infFile)  << inf.Added << "\t" << inf.Removed << "\t" << inf.Moved << "\t" << inf.Scanned/(double)(inf.Samples) << "\t" << inf.Activated/(double)(inf.Samples) << "\t" << UsedCapacity << endl;
  
  inf.Added=0;
  inf.Removed=0;
  inf.Moved=0;
  inf.Scanned=0;
  inf.Activated=0;
  inf.Samples=0;

  return UsedCapacity;

}

//=============================================


SDM::~SDM(){
  //cout << "Sutton Changes: " << SuttonChanges << endl;
  delete [] HardArray;

}

//-------------------------
void SDM::getSample(const State& s, State& sample){
  int i;

  for(i=0; i<State::dimensionality; i++){
    sample.x[i] = (double)rand()/(double)(RAND_MAX)*ActivationRadius[i]; 
  }

}

//------------------------------------------
//-----------------------------------------


sdmParameters::sdmParameters(){
  
  memCapacity = 0;
  dimensions = 0;
  radius = NULL;
  left = NULL;
  right = NULL;
  StartingSize=0;
  SufficientSamples=0;
  Threshold = -1.0;
  
}

//-----------------------------------------

void sdmParameters::processCommandLine(int argc, char* argv[]){

  int i, j, a;

  for (i=0; i<argc; i++){
    if (strncmp(argv[i],"?", 1)==0){

      cout << "SDMs command line parammeters:" << endl;
  
      cout << "dim= (number of dimensions)" << endl;
      cout << "ar= (activation radi(us/a); one value if the same for all dimensions or values separated by commas if different for each dimension)" << endl;
      cout << "lb= (left bound(s); one value if the same for all dimensions or values separated by commas if different for each dimension)" << endl;
      cout << "rb= (right bound(s); one value if the same for all dimensions or values separated by commas if different for each dimension)" << endl;
      cout << "hl= (a minimum desired number of active hard locations )" << endl;
      cout << "th= (close match threshold. set to -1 to be automatically determined from hl parameter)" << endl;
      cout << "mc= memory capacity" << endl;
      cout << "ss= starting size of the memory" << endl;
      cout << "alpha= (constant if using the constant schedule, or initial value if using decreasing schedules)" << endl;
      cout << "schedule= (possible values: constant, decrease)" << endl;
      cout << "d= (decrease factor, if using decrease schedule for alpha)" << endl;
      cout << "t= (trace usage option; possible values 1 (for yes) and 0 (for no))" << endl;
    
      return;
    }


    if (strncmp("mc=", argv[i], 3)==0)
      memCapacity=atoi(&(argv[i][3]));

    if (strncmp("ss=", argv[i], 3)==0)
      StartingSize=atoi(&(argv[i][3]));

    if (strncmp("hl=", argv[i], 3)==0)
      SufficientSamples=atoi(&(argv[i][3]));

    if (strncmp("dim=", argv[i], 4)==0)
      dimensions=atoi(&(argv[i][4]));

    if (strncmp("th=", argv[i], 3)==0)
      Threshold=atof(&(argv[i][3]));

    if (strncmp("ar=", argv[i], 3)==0){
      if (dimensions == 0){
	if (State::dimensionality == 0){
	  cout << "Error (SDMs): specify function dimensionality before the activation radia" << endl;
	  exit(EXIT_FAILURE);
	}
	else{
	  //cout << "Warning: using dimensionality = " << State::dimensionality << endl;
	  dimensions = State::dimensionality;

	}
	
      }
      radius = new double[dimensions];

	  
      if (strchr(argv[i],',')==NULL){
	for(j=0;j<dimensions;j++){
	  radius[j]=-1;
	  radius[j]=atof(&(argv[i][3]));
	  if (radius[j]<0) {
	    cout << "Error: radius Specified incorrectly" << endl;
	    exit(EXIT_FAILURE);
	  }
	}
      }
      else{
	a=tokenize(&(argv[i][3]), dimensions, radius);
	if (a==-1){
	  cout << "Error (SDMs): radius parameters specified incorrectly" << endl;
	  exit(EXIT_FAILURE);
	}
      }
    }


    if (strncmp("lb=", argv[i], 3)==0){
      if (dimensions == 0){
	if (State::dimensionality == 0){
	  cout << "Error (SDMs): specify function dimensionality before the bounds" << endl;
	  exit(EXIT_FAILURE);
	}
	else{
	  cout << "Warning: using dimensionality = " << State::dimensionality << endl;
	  dimensions = State::dimensionality;

	}
	
      }
      left = new double[dimensions];

	  
      if (strchr(argv[i],',')==NULL){
	for(j=0;j<dimensions;j++){
	  left[j]=-1;
	  left[j]=atof(&(argv[i][3]));
	  if (radius[j]<0) {
	    cout << "Error(SDMs) : left bound Specified incorrectly" << endl;
	    exit(EXIT_FAILURE);
	  }
	}
      }
      else{
	a=tokenize(&(argv[i][3]),dimensions, left);
	if (a==-1){
	  cout << "Error (SDMs): left bounds specified incorrectly" << endl;
	  exit(EXIT_FAILURE);
	}
      }
    }

    if (strncmp("rb=", argv[i], 3)==0){
      if (dimensions == 0){
	if (State::dimensionality == 0){
	  cout << "Error (SDMs): specify function dimensionality before the bounds" << endl;
	  exit(EXIT_FAILURE);
	}
	else{
	  cout << "Warning: using dimensionality = " << State::dimensionality << endl;
	  dimensions = State::dimensionality;

	}
	
      }
      right = new double[dimensions];

	  
      if (strchr(argv[i],',')==NULL){
	for(j=0;j<dimensions;j++){
	  right[j]=-1;
	  right[j]=atof(&(argv[i][3]));
	  if (radius[j]<0) {
	    cout << "Error(SDMs) : left bound Specified incorrectly" << endl;
	    exit(EXIT_FAILURE);
	  }
	}
      }
      else{
	a=tokenize(&(argv[i][3]), dimensions, right);
	if (a==-1){
	  cout << "Error (SDMs): left bounds specified incorrectly" << endl;
	  exit(EXIT_FAILURE);
	}
      }
	
	
    }
  }

  if (radius == NULL)
    radius = new double[dimensions];

  if (left == NULL)
    left = new double[dimensions];

  if (right == NULL)
    right = new double[dimensions];
  
  
      
}


    //--------------------------

int sdmParameters::tokenize(char* string, int n, double* tokens){
  char seps[]=",/";
  char* token=NULL;
  int i;

  //get first token
  token=strtok(string,seps);
  for (i=0; i<n; i++){
    if (token==NULL)
      return -1;
    tokens[i]=atof(token);
    token = strtok( NULL, seps );
  }
  return 1;
}

//-----------------------------------------

sdmParameters::~sdmParameters(){
  delete [] radius;

}

//------------------------------------------------------------
//-------------------------------------

HardLocation::HardLocation(){

  value=0;
  EligibilityTrace = 1;
  alpha = 0.01;
  EligibilityTrace = 0;
  ActivationCounter = 0;
  TimeStamp = 0;
}

//-----------------------------------------------------------


List::List(){

  head = NULL;
  walkPtr = NULL;

}

//---------

List::~List(){

  Node* ptr, *pn;
  ptr = head;

  while(ptr != NULL){
    
    head = head->next;
    delete ptr;
    ptr = head;
    
  }

}

//----------

List::List(Node* H){

  if (H == NULL) head = NULL;
  else{

    Node* ptrH = H;
    Node* ptr;
  
    head = new Node;
    head->prev = NULL;
    head->index = H->index;
    ptr = head;
  

    while(ptrH->next != NULL){
      ptr->next = new Node;
      ptrH = ptrH->next;
      ptr->next->prev = ptr;
      ptr->next->index = ptrH->index;
      ptr = ptr->next;
    }

    ptr->next = NULL;

  }
  walkPtr = NULL;
}

//----------

void List::add(int ind){
  
  Node* ptr = head;
  Node* newNode;

  if (head == NULL){
    head = new Node;
    head->prev = NULL;
    head->next = NULL;
    head->index = ind;
    return;
  }

  while (true){
    
    if (ptr->index > ind){
      if (ptr == head){
	head->prev = new Node;
	head->prev->prev = NULL;
	head->prev->next = head;
	head->prev->index = ind;
	head = head->prev;
      }
      else{
	newNode = new Node;
	newNode->index = ind;
	newNode->prev = ptr->prev;
	newNode->next = ptr;

	ptr->prev->next = newNode;
	ptr->prev = newNode;
	
	
      }
      
      return;

    }
    if (ptr->next != NULL)
      ptr = ptr->next;
    else{
      ptr->next = new Node;
      ptr->next->prev = ptr;
      ptr->next->index = ind;
      ptr->next->next = NULL;
      return;
    }

  }

}

//---------------

Node* List::find(int ind){

  Node* ptr = head;
  
  while (ptr != NULL){
    if (ptr->index == ind)
      return ptr;
    if (ptr->index > ind)
      return NULL;
    ptr = ptr->next;
  }
  return ptr;
}

//-------------

void List::remove(int ind){

  Node* toRemove = find(ind);
  if (toRemove == NULL){
    cout << "Error: Cannot remove HL " << ind << endl;
    exit(EXIT_FAILURE);
  }
  remove(toRemove);

}

//---------------------

void List::remove(Node* toRemove){

  
  if (toRemove == NULL) return;

  if (toRemove == head){
    if (head->next != NULL){
      head=head->next;
      head->prev = NULL;
    }
    else head=NULL;

    delete toRemove;
    return;
  }
  
  toRemove->prev->next = toRemove->next;
  if (toRemove->next != NULL)
    toRemove->next->prev = toRemove->prev;
  delete toRemove;


  
}

//------------------------

void List::listIntersection(List* L){

  

  if (head == NULL){

    return;
  }
  Node *l1, *l2, *l;
  l1 = head;
  l2 = L->head;

  while ((l1!=NULL) && (l2!=NULL)){

    if (l1->index == l2->index){
      l1 = l1->next;
      l2 = l2->next;
    }
    else{
      if (l1->index < l2->index){
	l=l1;
	l1 = l1->next;
	remove(l);
	
      }
      else{

	if (l1->index > l2->index){
	  l2 = l2->next;
	}
      }
    }

  }


  if (l1 != NULL){
    if (l1==head){
      delete head;
      head = NULL;
    }
    else{
      l1->prev->next = NULL;
      while (l1 != NULL){
	l=l1;
	l1=l1->next;
	delete l;
      }
    }
    
  }
} 

//------------------------

void List::listUnion(List* L){

  Node *l1, *l2;
  l1 = head;
  l2 = L->head;

  while ((l1!=NULL) && (l2!=NULL)){

    if (l1->index == l2->index){
      l1 = l1->next;
      l2 = l2->next;
      
    }
    else{

      if (l1->index < l2->index){ 
	l1 = l1->next;
      }
      else{

	if (l1->index > l2->index){
	  add(l2->index);
	  l2 = l2->next;
	}
      }
    }

  }

  while (l2 != NULL){
    
    add(l2->index);
    l2 = l2->next;
  }
}

//-------------------

Node* List::walk(bool mode){

  if (mode == start) walkPtr = head;
  if (mode == continue)
    if (walkPtr != NULL)
      walkPtr = walkPtr->next;
  return walkPtr;

}

/*
	Returns true iff the list is empty
*/
bool	List::isEmpty()
{
	if (head == NULL)
	{
		return true;
	}
	else
	{
		return false;
	}
}

//-----------------------
//-----------------------



//---------------------------------
//----------------------------------------------------------

