// ******************************************************
//  file    sdm.h
//  purpose class for sparse distributed memory FA
//  author  bohdana ratitch
//  written november 2003
// ******************************************************

#ifndef SDM_H
#define SDM_H

#include <stdio.h>
#include <assert.h>
#include <time.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>

#include "interface_classes.h"
#include "HLsupport.h"


#define DEFAULT_VALUE 0.0
#define RECENCY 0.95
#define DECREASE_ALPHA 1
#define CONSTANT_ALPHA 0
#define start 0
#define continue 1


class HardLocation{

 public:

  HardLocation();

  State state;
  double value;
  double alpha;
  double EligibilityTrace;
  double ActivationCounter;
  unsigned int TimeStamp;
     
};


struct MemoryInfo{

  int Added;
  int Removed;
  int Moved;
  int Activated;
  int Scanned;
  int Samples;

};

struct Node{
 
  int index;
  Node* prev;
  Node* next;

};

class List{

 public:

  List();
  List(Node* H);
  void add(int ind);
  void remove(int ind);
  void remove(Node* item);
  Node* find(int ind);
  ~List();
  Node* walk(bool mode);
  void listIntersection(List* L);
  void listUnion(List* L);
  
  bool isEmpty();
  

 protected:
 
  Node* head;
  Node* walkPtr;

};

struct sdmParameters{

  int memCapacity;
  double* radius;
  double* left;
  double* right;
  int dimensions;
  int StartingSize;
  int SufficientSamples;
  double Threshold;

  sdmParameters();
  void processCommandLine(int argc, char* argv[]);
  int tokenize(char* string, int n, double* tokens);
  ~sdmParameters();

};

struct NeighborhoodInfo{

  int Activated;
  double MaxWeight;
  int MaxIndex;
  int NumClose;
  double WeightSum;
  double Prediction;
  int RandIndex;
  double RandWeight;
};

//----------------------------------------------------------
//-----------------------------------------------------------


class SDM : public Approximator {


 public:

  SDM(sdmParameters& sdmP, double (*simFunction)(const State& s1, const State& s2, double* threshold));

  void predict(const State& s, double& output);
  

  void learn(const State& s, const double target);
  /* Learns an input-output pair.
     s : input (state)
     target: target output value
  */

  void computeGradient(const State& s, double* GradientVector);
  /* Compute the gradient w.r.t. architecture parameters at the current parameters' values and input s
   */

  void updateParameters(double* delta);
  /* Update parameters by amounts in delta array, 
     possible multiplied by appropriate learning steps.
  */

  void replaceTraces(const State& s, double replace);
  /* Replace traces of parameters, activated by input state s to value replace.
   */

  void decayTraces(double factor);
  /* Decay (multiply) traces by factor.
   */

  void accumulateTraces(const State& s, double amount);
  /* Increment traces by amount for parameters activated by input s.
   */

  void setArchitectureParameters(int argc, char *argv[]);
  /* Loads parameters of the architecture from a text file.
     argc : number of supplied arguments
     argv : array of arguments
  */

  void saveArchitectureParameters(int argc, char *argv[]);
  /* Saves parameters of the architecture into a text file.
     argc : number of supplied arguments
     argv : array of arguments
  */

  void setLearningParameters(int argc, char *argv[]);
  /* Sets learning parameters (e.g. learning step).
     argc : number of supplied arguments
     argv : array of arguments
  */


  int getSize();
  /* Return the number of parameters in this architecture
   */

  void setRehearsing(bool mode);

  void setID(int id);

  

  ~SDM();

  

protected:

  double (*similarity)(const State& s1, const State& s2, double* threshold);
  void getSample(const State& s, State& sample);
  void learnNoTraces(const State& s, double target);
  void learnWithTraces(const State& s, double target);
  void randomSample(const State& s, int Attempts, State& sNew, bool& success, double& output);
  void randomOutSample(const State& s, int Attempts, State& sample, bool& success);
  double localPredict(const State& s);
  void localActive(const State& s, NeighborhoodInfo& ni);
  void localUpdate(const State& s, double target);
  void removeHL(int index);
  bool removeHL(State& s, NeighborhoodInfo& ni);
  void addHL(const State& s, NeighborhoodInfo& ni);
  

  //=======data=============

  unsigned int Capacity;
  unsigned int UsedCapacity;
  HardLocation* HardArray;
  double* ActivationRadius;
  unsigned int SufficientSamples;
  double InitialAlpha;
  bool AlphaSchedule;
  double AlphaDecreaseConst;
  HLsupport* support;
  bool UseTraces;
  unsigned int Time;
  double Rehearsing;
  double RememberRehearsing;
  double Threshold;
  int RandomAdditionsFactor;
  int GetInactiveAttempt;
  int MaximumSamples;
  int SuttonChanges;
  MemoryInfo inf;
  int ID;
  ofstream* infFile;
   
};


#endif

